"""UUID utilities."""
from __future__ import annotations

from typing import Callable
from uuid import UUID, uuid4


def uuid(_uuid: Callable[[], UUID] = uuid4) -> str:
    """Generate unique id in UUID4 format.

    See Also
    --------
        For now this is provided by :func:`uuid.uuid4`.
    """
    return str(_uuid())



"""SoftLayer Message Queue transport module for kombu.

Features
========
* Type: Virtual
* Supports Direct: Yes
* Supports Topic: Yes
* Supports Fanout: No
* Supports Priority: No
* Supports TTL: No

Connection String
=================
 *Unreviewed*

Transport Options
=================
 *Unreviewed*
"""

from __future__ import annotations

import os
import socket
import string
from queue import Empty

from kombu.utils.encoding import bytes_to_str, safe_str
from kombu.utils.json import dumps, loads
from kombu.utils.objects import cached_property

from . import virtual

try:
    from softlayer_messaging import get_client
    from softlayer_messaging.errors import ResponseError
except ImportError:  # pragma: no cover
    get_client = ResponseError = None

# dots are replaced by dash, all other punctuation replaced by underscore.
CHARS_REPLACE_TABLE = {
    ord(c): 0x5f for c in string.punctuation if c not in '_'
}


class Channel(virtual.Channel):
    """SLMQ Channel."""

    default_visibility_timeout = 1800  # 30 minutes.
    domain_format = 'kombu%(vhost)s'
    _slmq = None
    _queue_cache = {}
    _noack_queues = set()

    def __init__(self, *args, **kwargs):
        if get_client is None:
            raise ImportError(
                'SLMQ transport requires the softlayer_messaging library',
            )
        super().__init__(*args, **kwargs)
        queues = self.slmq.queues()
        for queue in queues:
            self._queue_cache[queue] = queue

    def basic_consume(self, queue, no_ack, *args, **kwargs):
        if no_ack:
            self._noack_queues.add(queue)
        return super().basic_consume(queue, no_ack,
                                     *args, **kwargs)

    def basic_cancel(self, consumer_tag):
        if consumer_tag in self._consumers:
            queue = self._tag_to_queue[consumer_tag]
            self._noack_queues.discard(queue)
        return super().basic_cancel(consumer_tag)

    def entity_name(self, name, table=CHARS_REPLACE_TABLE):
        """Format AMQP queue name into a valid SLQS queue name."""
        return str(safe_str(name)).translate(table)

    def _new_queue(self, queue, **kwargs):
        """Ensure a queue exists in SLQS."""
        queue = self.entity_name(self.queue_name_prefix + queue)
        try:
            return self._queue_cache[queue]
        except KeyError:
            try:
                self.slmq.create_queue(
                    queue, visibility_timeout=self.visibility_timeout)
            except ResponseError:
                pass
            q = self._queue_cache[queue] = self.slmq.queue(queue)
            return q

    def _delete(self, queue, *args, **kwargs):
        """Delete queue by name."""
        queue_name = self.entity_name(queue)
        self._queue_cache.pop(queue_name, None)
        self.slmq.queue(queue_name).delete(force=True)
        super()._delete(queue_name)

    def _put(self, queue, message, **kwargs):
        """Put message onto queue."""
        q = self._new_queue(queue)
        q.push(dumps(message))

    def _get(self, queue):
        """Try to retrieve a single message off ``queue``."""
        q = self._new_queue(queue)
        rs = q.pop(1)
        if rs['items']:
            m = rs['items'][0]
            payload = loads(bytes_to_str(m['body']))
            if queue in self._noack_queues:
                q.message(m['id']).delete()
            else:
                payload['properties']['delivery_info'].update({
                    'slmq_message_id': m['id'], 'slmq_queue_name': q.name})
            return payload
        raise Empty()

    def basic_ack(self, delivery_tag):
        delivery_info = self.qos.get(delivery_tag).delivery_info
        try:
            queue = delivery_info['slmq_queue_name']
        except KeyError:
            pass
        else:
            self.delete_message(queue, delivery_info['slmq_message_id'])
        super().basic_ack(delivery_tag)

    def _size(self, queue):
        """Return the number of messages in a queue."""
        return self._new_queue(queue).detail()['message_count']

    def _purge(self, queue):
        """Delete all current messages in a queue."""
        q = self._new_queue(queue)
        n = 0
        results = q.pop(10)
        while results['items']:
            for m in results['items']:
                self.delete_message(queue, m['id'])
                n += 1
            results = q.pop(10)
        return n

    def delete_message(self, queue, message_id):
        q = self.slmq.queue(self.entity_name(queue))
        return q.message(message_id).delete()

    @property
    def slmq(self):
        if self._slmq is None:
            conninfo = self.conninfo
            account = os.environ.get('SLMQ_ACCOUNT', conninfo.virtual_host)
            user = os.environ.get('SL_USERNAME', conninfo.userid)
            api_key = os.environ.get('SL_API_KEY', conninfo.password)
            host = os.environ.get('SLMQ_HOST', conninfo.hostname)
            port = os.environ.get('SLMQ_PORT', conninfo.port)
            secure = bool(os.environ.get(
                'SLMQ_SECURE', self.transport_options.get('secure')) or True,
            )
            endpoint = '{}://{}{}'.format(
                'https' if secure else 'http', host,
                f':{port}' if port else '',
            )

            self._slmq = get_client(account, endpoint=endpoint)
            self._slmq.authenticate(user, api_key)
        return self._slmq

    @property
    def conninfo(self):
        return self.connection.client

    @property
    def transport_options(self):
        return self.connection.client.transport_options

    @cached_property
    def visibility_timeout(self):
        return (self.transport_options.get('visibility_timeout') or
                self.default_visibility_timeout)

    @cached_property
    def queue_name_prefix(self):
        return self.transport_options.get('queue_name_prefix', '')


class Transport(virtual.Transport):
    """SLMQ Transport."""

    Channel = Channel

    polling_interval = 1
    default_port = None
    connection_errors = (
        virtual.Transport.connection_errors + (
            ResponseError, socket.error
        )
    )



"""Amazon SQS transport module for Kombu.

This package implements an AMQP-like interface on top of Amazons SQS service,
with the goal of being optimized for high performance and reliability.

The default settings for this module are focused now on high performance in
task queue situations where tasks are small, idempotent and run very fast.

SQS Features supported by this transport
========================================
Long Polling
------------
https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html

Long polling is enabled by setting the `wait_time_seconds` transport
option to a number > 1.  Amazon supports up to 20 seconds.  This is
enabled with 10 seconds by default.

Batch API Actions
-----------------
https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-batch-api.html

The default behavior of the SQS Channel.drain_events() method is to
request up to the 'prefetch_count' messages on every request to SQS.
These messages are stored locally in a deque object and passed back
to the Transport until the deque is empty, before triggering a new
API call to Amazon.

This behavior dramatically speeds up the rate that you can pull tasks
from SQS when you have short-running tasks (or a large number of workers).

When a Celery worker has multiple queues to monitor, it will pull down
up to 'prefetch_count' messages from queueA and work on them all before
moving on to queueB.  If queueB is empty, it will wait up until
'polling_interval' expires before moving back and checking on queueA.

Message Attributes
-----------------
https://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-message-metadata.html

SQS supports sending message attributes along with the message body.
To use this feature, you can pass a 'message_attributes' as keyword argument
to `basic_publish` method.

Other Features supported by this transport
==========================================
Predefined Queues
-----------------
The default behavior of this transport is to use a single AWS credential
pair in order to manage all SQS queues (e.g. listing queues, creating
queues, polling queues, deleting messages).

If it is preferable for your environment to use multiple AWS credentials, you
can use the 'predefined_queues' setting inside the 'transport_options' map.
This setting allows you to specify the SQS queue URL and AWS credentials for
each of your queues. For example, if you have two queues which both already
exist in AWS) you can tell this transport about them as follows:

.. code-block:: python

    transport_options = {
      'predefined_queues': {
        'queue-1': {
          'url': 'https://sqs.us-east-1.amazonaws.com/xxx/aaa',
          'access_key_id': 'a',
          'secret_access_key': 'b',
          'backoff_policy': {1: 10, 2: 20, 3: 40, 4: 80, 5: 320, 6: 640}, # optional
          'backoff_tasks': ['svc.tasks.tasks.task1'] # optional
        },
        'queue-2.fifo': {
          'url': 'https://sqs.us-east-1.amazonaws.com/xxx/bbb.fifo',
          'access_key_id': 'c',
          'secret_access_key': 'd',
          'backoff_policy': {1: 10, 2: 20, 3: 40, 4: 80, 5: 320, 6: 640}, # optional
          'backoff_tasks': ['svc.tasks.tasks.task2'] # optional
        },
      }
    'sts_role_arn': 'arn:aws:iam::<xxx>:role/STSTest', # optional
    'sts_token_timeout': 900 # optional
    }

Note that FIFO and standard queues must be named accordingly (the name of
a FIFO queue must end with the .fifo suffix).

backoff_policy & backoff_tasks are optional arguments. These arguments
automatically change the message visibility timeout, in order to have
different times between specific task retries. This would apply after
task failure.

AWS STS authentication is supported, by using sts_role_arn, and
sts_token_timeout. sts_role_arn is the assumed IAM role ARN we are trying
to access with. sts_token_timeout is the token timeout, defaults (and minimum)
to 900 seconds. After the mentioned period, a new token will be created.



If you authenticate using Okta_ (e.g. calling |gac|_), you can also specify
a 'session_token' to connect to a queue. Note that those tokens have a
limited lifetime and are therefore only suited for short-lived tests.

.. _Okta: https://www.okta.com/
.. _gac: https://github.com/Nike-Inc/gimme-aws-creds#readme
.. |gac| replace:: ``gimme-aws-creds``


Client config
-------------
In some cases you may need to override the botocore config. You can do it
as follows:

.. code-block:: python

    transport_option = {
      'client-config': {
          'connect_timeout': 5,
       },
    }

For a complete list of settings you can adjust using this option see
https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html

Features
========
* Type: Virtual
* Supports Direct: Yes
* Supports Topic: Yes
* Supports Fanout: Yes
* Supports Priority: No
* Supports TTL: No
"""


from __future__ import annotations

import base64
import socket
import string
import uuid
from datetime import datetime
from queue import Empty

from botocore.client import Config
from botocore.exceptions import ClientError
from vine import ensure_promise, promise, transform

from kombu.asynchronous import get_event_loop
from kombu.asynchronous.aws.ext import boto3, exceptions
from kombu.asynchronous.aws.sqs.connection import AsyncSQSConnection
from kombu.asynchronous.aws.sqs.message import AsyncMessage
from kombu.log import get_logger
from kombu.utils import scheduling
from kombu.utils.encoding import bytes_to_str, safe_str
from kombu.utils.json import dumps, loads
from kombu.utils.objects import cached_property

from . import virtual

logger = get_logger(__name__)

# dots are replaced by dash, dash remains dash, all other punctuation
# replaced by underscore.
CHARS_REPLACE_TABLE = {
    ord(c): 0x5f for c in string.punctuation if c not in '-_.'
}
CHARS_REPLACE_TABLE[0x2e] = 0x2d  # '.' -> '-'

#: SQS bulk get supports a maximum of 10 messages at a time.
SQS_MAX_MESSAGES = 10


def maybe_int(x):
    """Try to convert x' to int, or return x' if that fails."""
    try:
        return int(x)
    except ValueError:
        return x


class UndefinedQueueException(Exception):
    """Predefined queues are being used and an undefined queue was used."""


class InvalidQueueException(Exception):
    """Predefined queues are being used and configuration is not valid."""


class AccessDeniedQueueException(Exception):
    """Raised when access to the AWS queue is denied.

    This may occur if the permissions are not correctly set or the
    credentials are invalid.
    """


class DoesNotExistQueueException(Exception):
    """The specified queue doesn't exist."""


class QoS(virtual.QoS):
    """Quality of Service guarantees implementation for SQS."""

    def reject(self, delivery_tag, requeue=False):
        super().reject(delivery_tag, requeue=requeue)
        routing_key, message, backoff_tasks, backoff_policy = \
            self._extract_backoff_policy_configuration_and_message(
                delivery_tag)
        if routing_key and message and backoff_tasks and backoff_policy:
            self.apply_backoff_policy(
                routing_key, delivery_tag, backoff_policy, backoff_tasks)

    def _extract_backoff_policy_configuration_and_message(self, delivery_tag):
        try:
            message = self._delivered[delivery_tag]
            routing_key = message.delivery_info['routing_key']
        except KeyError:
            return None, None, None, None
        if not routing_key or not message:
            return None, None, None, None
        queue_config = self.channel.predefined_queues.get(routing_key, {})
        backoff_tasks = queue_config.get('backoff_tasks')
        backoff_policy = queue_config.get('backoff_policy')
        return routing_key, message, backoff_tasks, backoff_policy

    def apply_backoff_policy(self, routing_key, delivery_tag,
                             backoff_policy, backoff_tasks):
        queue_url = self.channel._queue_cache[routing_key]
        task_name, number_of_retries = \
            self.extract_task_name_and_number_of_retries(delivery_tag)
        if not task_name or not number_of_retries:
            return None
        policy_value = backoff_policy.get(number_of_retries)
        if task_name in backoff_tasks and policy_value is not None:
            c = self.channel.sqs(routing_key)
            c.change_message_visibility(
                QueueUrl=queue_url,
                ReceiptHandle=delivery_tag,
                VisibilityTimeout=policy_value
            )

    def extract_task_name_and_number_of_retries(self, delivery_tag):
        message = self._delivered[delivery_tag]
        message_headers = message.headers
        task_name = message_headers['task']
        number_of_retries = int(
            message.properties['delivery_info']['sqs_message']
                              ['Attributes']['ApproximateReceiveCount'])
        return task_name, number_of_retries


class Channel(virtual.Channel):
    """SQS Channel."""

    default_region = 'us-east-1'
    default_visibility_timeout = 1800  # 30 minutes.
    default_wait_time_seconds = 10  # up to 20 seconds max
    domain_format = 'kombu%(vhost)s'
    _asynsqs = None
    _predefined_queue_async_clients = {}  # A client for each predefined queue
    _sqs = None
    _predefined_queue_clients = {}  # A client for each predefined queue
    _queue_cache = {}  # SQS queue name => SQS queue URL
    _noack_queues = set()
    QoS = QoS

    def __init__(self, *args, **kwargs):
        if boto3 is None:
            raise ImportError('boto3 is not installed')
        super().__init__(*args, **kwargs)
        self._validate_predifined_queues()

        # SQS blows up if you try to create a new queue when one already
        # exists but with a different visibility_timeout.  This prepopulates
        # the queue_cache to protect us from recreating
        # queues that are known to already exist.
        self._update_queue_cache(self.queue_name_prefix)

        self.hub = kwargs.get('hub') or get_event_loop()

    def _validate_predifined_queues(self):
        """Check that standard and FIFO queues are named properly.

        AWS requires FIFO queues to have a name
        that ends with the .fifo suffix.
        """
        for queue_name, q in self.predefined_queues.items():
            fifo_url = q['url'].endswith('.fifo')
            fifo_name = queue_name.endswith('.fifo')
            if fifo_url and not fifo_name:
                raise InvalidQueueException(
                    "Queue with url '{}' must have a name "
                    "ending with .fifo".format(q['url'])
                )
            elif not fifo_url and fifo_name:
                raise InvalidQueueException(
                    "Queue with name '{}' is not a FIFO queue: "
                    "'{}'".format(queue_name, q['url'])
                )

    def _update_queue_cache(self, queue_name_prefix):
        if self.predefined_queues:
            for queue_name, q in self.predefined_queues.items():
                self._queue_cache[queue_name] = q['url']
            return

        resp = self.sqs().list_queues(QueueNamePrefix=queue_name_prefix)
        for url in resp.get('QueueUrls', []):
            queue_name = url.split('/')[-1]
            self._queue_cache[queue_name] = url

    def basic_consume(self, queue, no_ack, *args, **kwargs):
        if no_ack:
            self._noack_queues.add(queue)
        if self.hub:
            self._loop1(queue)
        return super().basic_consume(
            queue, no_ack, *args, **kwargs
        )

    def basic_cancel(self, consumer_tag):
        if consumer_tag in self._consumers:
            queue = self._tag_to_queue[consumer_tag]
            self._noack_queues.discard(queue)
        return super().basic_cancel(consumer_tag)

    def drain_events(self, timeout=None, callback=None, **kwargs):
        """Return a single payload message from one of our queues.

        Raises
        ------
            Queue.Empty: if no messages available.
        """
        # If we're not allowed to consume or have no consumers, raise Empty
        if not self._consumers or not self.qos.can_consume():
            raise Empty()

        # At this point, go and get more messages from SQS
        self._poll(self.cycle, callback, timeout=timeout)

    def _reset_cycle(self):
        """Reset the consume cycle.

        Returns
        -------
            FairCycle: object that points to our _get_bulk() method
                rather than the standard _get() method.  This allows for
                multiple messages to be returned at once from SQS (
                based on the prefetch limit).
        """
        self._cycle = scheduling.FairCycle(
            self._get_bulk, self._active_queues, Empty,
        )

    def entity_name(self, name, table=CHARS_REPLACE_TABLE):
        """Format AMQP queue name into a legal SQS queue name."""
        if name.endswith('.fifo'):
            partial = name[:-len('.fifo')]
            partial = str(safe_str(partial)).translate(table)
            return partial + '.fifo'
        else:
            return str(safe_str(name)).translate(table)

    def canonical_queue_name(self, queue_name):
        return self.entity_name(self.queue_name_prefix + queue_name)

    def _resolve_queue_url(self, queue):
        """Try to retrieve the SQS queue URL for a given queue name."""
        # Translate to SQS name for consistency with initial
        # _queue_cache population.
        sqs_qname = self.canonical_queue_name(queue)

        # The SQS ListQueues method only returns 1000 queues.  When you have
        # so many queues, it's possible that the queue you are looking for is
        # not cached.  In this case, we could update the cache with the exact
        # queue name first.
        if sqs_qname not in self._queue_cache:
            self._update_queue_cache(sqs_qname)
        try:
            return self._queue_cache[sqs_qname]
        except KeyError:
            if self.predefined_queues:
                raise UndefinedQueueException((
                    "Queue with name '{}' must be "
                    "defined in 'predefined_queues'."
                ).format(sqs_qname))

            raise DoesNotExistQueueException(
                f"Queue with name '{sqs_qname}' doesn't exist in SQS"
            )

    def _new_queue(self, queue, **kwargs):
        """Ensure a queue with given name exists in SQS.

        Arguments:
        ---------
            queue (str): the AMQP queue name
        Returns
            str: the SQS queue URL
        """
        try:
            return self._resolve_queue_url(queue)
        except DoesNotExistQueueException:
            sqs_qname = self.canonical_queue_name(queue)
            attributes = {'VisibilityTimeout': str(self.visibility_timeout)}
            if sqs_qname.endswith('.fifo'):
                attributes['FifoQueue'] = 'true'

            resp = self._create_queue(sqs_qname, attributes)
            self._queue_cache[sqs_qname] = resp['QueueUrl']
            return resp['QueueUrl']

    def _create_queue(self, queue_name, attributes):
        """Create an SQS queue with a given name and nominal attributes."""
        # Allow specifying additional boto create_queue Attributes
        # via transport options
        if self.predefined_queues:
            return None

        attributes.update(
            self.transport_options.get('sqs-creation-attributes') or {},
        )

        return self.sqs(queue=queue_name).create_queue(
            QueueName=queue_name,
            Attributes=attributes,
        )

    def _delete(self, queue, *args, **kwargs):
        """Delete queue by name."""
        if self.predefined_queues:
            return

        q_url = self._resolve_queue_url(queue)
        self.sqs().delete_queue(
            QueueUrl=q_url,
        )
        self._queue_cache.pop(queue, None)

    def _put(self, queue, message, **kwargs):
        """Put message onto queue."""
        q_url = self._new_queue(queue)
        kwargs = {'QueueUrl': q_url}
        if 'properties' in message:
            if 'message_attributes' in message['properties']:
                # we don't want to want to have the attribute in the body
                kwargs['MessageAttributes'] = \
                    message['properties'].pop('message_attributes')
            if queue.endswith('.fifo'):
                if 'MessageGroupId' in message['properties']:
                    kwargs['MessageGroupId'] = \
                        message['properties']['MessageGroupId']
                else:
                    kwargs['MessageGroupId'] = 'default'
                if 'MessageDeduplicationId' in message['properties']:
                    kwargs['MessageDeduplicationId'] = \
                        message['properties']['MessageDeduplicationId']
                else:
                    kwargs['MessageDeduplicationId'] = str(uuid.uuid4())
            else:
                if "DelaySeconds" in message['properties']:
                    kwargs['DelaySeconds'] = \
                        message['properties']['DelaySeconds']

        if self.sqs_base64_encoding:
            body = AsyncMessage().encode(dumps(message))
        else:
            body = dumps(message)
        kwargs['MessageBody'] = body

        c = self.sqs(queue=self.canonical_queue_name(queue))
        if message.get('redelivered'):
            c.change_message_visibility(
                QueueUrl=q_url,
                ReceiptHandle=message['properties']['delivery_tag'],
                VisibilityTimeout=0
            )
        else:
            c.send_message(**kwargs)

    @staticmethod
    def _optional_b64_decode(byte_string):
        try:
            data = base64.b64decode(byte_string)
            if base64.b64encode(data) == byte_string:
                return data
            # else the base64 module found some embedded base64 content
            # that should be ignored.
        except Exception:  # pylint: disable=broad-except
            pass
        return byte_string

    def _message_to_python(self, message, queue_name, q_url):
        body = self._optional_b64_decode(message['Body'].encode())
        payload = loads(bytes_to_str(body))
        if queue_name in self._noack_queues:
            q_url = self._new_queue(queue_name)
            self.asynsqs(queue=queue_name).delete_message(
                q_url,
                message['ReceiptHandle'],
            )
        else:
            try:
                properties = payload['properties']
                delivery_info = payload['properties']['delivery_info']
            except KeyError:
                # json message not sent by kombu?
                delivery_info = {}
                properties = {'delivery_info': delivery_info}
                payload.update({
                    'body': bytes_to_str(body),
                    'properties': properties,
                })
            # set delivery tag to SQS receipt handle
            delivery_info.update({
                'sqs_message': message, 'sqs_queue': q_url,
            })
            properties['delivery_tag'] = message['ReceiptHandle']
        return payload

    def _messages_to_python(self, messages, queue):
        """Convert a list of SQS Message objects into Payloads.

        This method handles converting SQS Message objects into
        Payloads, and appropriately updating the queue depending on
        the 'ack' settings for that queue.

        Arguments:
        ---------
            messages (SQSMessage): A list of SQS Message objects.
            queue (str): Name representing the queue they came from.

        Returns
        -------
            List: A list of Payload objects
        """
        q_url = self._new_queue(queue)
        return [self._message_to_python(m, queue, q_url) for m in messages]

    def _get_bulk(self, queue,
                  max_if_unlimited=SQS_MAX_MESSAGES, callback=None):
        """Try to retrieve multiple messages off ``queue``.

        Where :meth:`_get` returns a single Payload object, this method
        returns a list of Payload objects.  The number of objects returned
        is determined by the total number of messages available in the queue
        and the number of messages the QoS object allows (based on the
        prefetch_count).

        Note:
        ----
            Ignores QoS limits so caller is responsible for checking
            that we are allowed to consume at least one message from the
            queue.  get_bulk will then ask QoS for an estimate of
            the number of extra messages that we can consume.

        Arguments:
        ---------
            queue (str): The queue name to pull from.

        Returns
        -------
            List[Message]
        """
        # drain_events calls `can_consume` first, consuming
        # a token, so we know that we are allowed to consume at least
        # one message.

        # Note: ignoring max_messages for SQS with boto3
        max_count = self._get_message_estimate()
        if max_count:
            q_url = self._new_queue(queue)
            resp = self.sqs(queue=queue).receive_message(
                QueueUrl=q_url, MaxNumberOfMessages=max_count,
                WaitTimeSeconds=self.wait_time_seconds)
            if resp.get('Messages'):
                for m in resp['Messages']:
                    m['Body'] = AsyncMessage(body=m['Body']).decode()
                for msg in self._messages_to_python(resp['Messages'], queue):
                    self.connection._deliver(msg, queue)
                return
        raise Empty()

    def _get(self, queue):
        """Try to retrieve a single message off ``queue``."""
        q_url = self._new_queue(queue)
        resp = self.sqs(queue=queue).receive_message(
            QueueUrl=q_url, MaxNumberOfMessages=1,
            WaitTimeSeconds=self.wait_time_seconds)
        if resp.get('Messages'):
            body = AsyncMessage(body=resp['Messages'][0]['Body']).decode()
            resp['Messages'][0]['Body'] = body
            return self._messages_to_python(resp['Messages'], queue)[0]
        raise Empty()

    def _loop1(self, queue, _=None):
        self.hub.call_soon(self._schedule_queue, queue)

    def _schedule_queue(self, queue):
        if queue in self._active_queues:
            if self.qos.can_consume():
                self._get_bulk_async(
                    queue, callback=promise(self._loop1, (queue,)),
                )
            else:
                self._loop1(queue)

    def _get_message_estimate(self, max_if_unlimited=SQS_MAX_MESSAGES):
        maxcount = self.qos.can_consume_max_estimate()
        return min(
            max_if_unlimited if maxcount is None else max(maxcount, 1),
            max_if_unlimited,
        )

    def _get_bulk_async(self, queue,
                        max_if_unlimited=SQS_MAX_MESSAGES, callback=None):
        maxcount = self._get_message_estimate()
        if maxcount:
            return self._get_async(queue, maxcount, callback=callback)
        # Not allowed to consume, make sure to notify callback..
        callback = ensure_promise(callback)
        callback([])
        return callback

    def _get_async(self, queue, count=1, callback=None):
        q_url = self._new_queue(queue)
        qname = self.canonical_queue_name(queue)
        return self._get_from_sqs(
            queue_name=qname, queue_url=q_url, count=count,
            connection=self.asynsqs(queue=qname),
            callback=transform(
                self._on_messages_ready, callback, q_url, queue
            ),
        )

    def _on_messages_ready(self, queue, qname, messages):
        if 'Messages' in messages and messages['Messages']:
            callbacks = self.connection._callbacks
            for msg in messages['Messages']:
                msg_parsed = self._message_to_python(msg, qname, queue)
                callbacks[qname](msg_parsed)

    def _get_from_sqs(self, queue_name, queue_url,
                      connection, count=1, callback=None):
        """Retrieve and handle messages from SQS.

        Uses long polling and returns :class:`~vine.promises.promise`.
        """
        return connection.receive_message(
            queue_name, queue_url, number_messages=count,
            wait_time_seconds=self.wait_time_seconds,
            callback=callback,
        )

    def _restore(self, message,
                 unwanted_delivery_info=('sqs_message', 'sqs_queue')):
        for unwanted_key in unwanted_delivery_info:
            # Remove objects that aren't JSON serializable (Issue #1108).
            message.delivery_info.pop(unwanted_key, None)
        return super()._restore(message)

    def basic_ack(self, delivery_tag, multiple=False):
        try:
            message = self.qos.get(delivery_tag).delivery_info
            sqs_message = message['sqs_message']
        except KeyError:
            super().basic_ack(delivery_tag)
        else:
            queue = None
            if 'routing_key' in message:
                queue = self.canonical_queue_name(message['routing_key'])

            try:
                self.sqs(queue=queue).delete_message(
                    QueueUrl=message['sqs_queue'],
                    ReceiptHandle=sqs_message['ReceiptHandle']
                )
            except ClientError as exception:
                if exception.response['Error']['Code'] == 'AccessDenied':
                    raise AccessDeniedQueueException(
                        exception.response["Error"]["Message"]
                        )
                super().basic_reject(delivery_tag)
            else:
                super().basic_ack(delivery_tag)

    def _size(self, queue):
        """Return the number of messages in a queue."""
        q_url = self._new_queue(queue)
        c = self.sqs(queue=self.canonical_queue_name(queue))
        resp = c.get_queue_attributes(
            QueueUrl=q_url,
            AttributeNames=['ApproximateNumberOfMessages'])
        return int(resp['Attributes']['ApproximateNumberOfMessages'])

    def _purge(self, queue):
        """Delete all current messages in a queue."""
        q_url = self._new_queue(queue)
        # SQS is slow at registering messages, so run for a few
        # iterations to ensure messages are detected and deleted.
        size = 0
        for i in range(10):
            size += int(self._size(queue))
            if not size:
                break
        self.sqs(queue=queue).purge_queue(QueueUrl=q_url)
        return size

    def close(self):
        super().close()
        # if self._asynsqs:
        #     try:
        #         self.asynsqs().close()
        #     except AttributeError as exc:  # FIXME ???
        #         if "can't set attribute" not in str(exc):
        #             raise

    def new_sqs_client(self, region, access_key_id,
                       secret_access_key, session_token=None):
        session = boto3.session.Session(
            region_name=region,
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            aws_session_token=session_token,
        )
        is_secure = self.is_secure if self.is_secure is not None else True
        client_kwargs = {
            'use_ssl': is_secure
        }
        if self.endpoint_url is not None:
            client_kwargs['endpoint_url'] = self.endpoint_url
        client_config = self.transport_options.get('client-config') or {}
        config = Config(**client_config)
        return session.client('sqs', config=config, **client_kwargs)

    def sqs(self, queue=None):
        if queue is not None and self.predefined_queues:

            if queue not in self.predefined_queues:
                raise UndefinedQueueException(
                    f"Queue with name '{queue}' must be defined"
                    " in 'predefined_queues'.")
            q = self.predefined_queues[queue]
            if self.transport_options.get('sts_role_arn'):
                return self._handle_sts_session(queue, q)
            if not self.transport_options.get('sts_role_arn'):
                if queue in self._predefined_queue_clients:
                    return self._predefined_queue_clients[queue]
                else:
                    c = self._predefined_queue_clients[queue] = \
                        self.new_sqs_client(
                            region=q.get('region', self.region),
                            access_key_id=q.get(
                                'access_key_id', self.conninfo.userid),
                            secret_access_key=q.get(
                                'secret_access_key', self.conninfo.password)
                    )
                    return c

        if self._sqs is not None:
            return self._sqs

        c = self._sqs = self.new_sqs_client(
            region=self.region,
            access_key_id=self.conninfo.userid,
            secret_access_key=self.conninfo.password,
        )
        return c

    def _handle_sts_session(self, queue, q):
        region = q.get('region', self.region)
        if not hasattr(self, 'sts_expiration'):  # STS token - token init
            return self._new_predefined_queue_client_with_sts_session(queue, region)
        # STS token - refresh if expired
        elif self.sts_expiration.replace(tzinfo=None) < datetime.utcnow():
            return self._new_predefined_queue_client_with_sts_session(queue, region)
        else:  # STS token - ruse existing
            if queue not in self._predefined_queue_clients:
                return self._new_predefined_queue_client_with_sts_session(queue, region)
            return self._predefined_queue_clients[queue]

    def _new_predefined_queue_client_with_sts_session(self, queue, region):
        sts_creds = self.generate_sts_session_token(
            self.transport_options.get('sts_role_arn'),
            self.transport_options.get('sts_token_timeout', 900))
        self.sts_expiration = sts_creds['Expiration']
        c = self._predefined_queue_clients[queue] = self.new_sqs_client(
            region=region,
            access_key_id=sts_creds['AccessKeyId'],
            secret_access_key=sts_creds['SecretAccessKey'],
            session_token=sts_creds['SessionToken'],
        )
        return c

    def generate_sts_session_token(self, role_arn, token_expiry_seconds):
        sts_client = boto3.client('sts')
        sts_policy = sts_client.assume_role(
            RoleArn=role_arn,
            RoleSessionName='Celery',
            DurationSeconds=token_expiry_seconds
        )
        return sts_policy['Credentials']

    def asynsqs(self, queue=None):
        if queue is not None and self.predefined_queues:
            if queue in self._predefined_queue_async_clients and \
               not hasattr(self, 'sts_expiration'):
                return self._predefined_queue_async_clients[queue]
            if queue not in self.predefined_queues:
                raise UndefinedQueueException((
                    "Queue with name '{}' must be defined in "
                    "'predefined_queues'."
                ).format(queue))
            q = self.predefined_queues[queue]
            c = self._predefined_queue_async_clients[queue] = \
                AsyncSQSConnection(
                    sqs_connection=self.sqs(queue=queue),
                    region=q.get('region', self.region),
                    fetch_message_attributes=self.fetch_message_attributes,
            )
            return c

        if self._asynsqs is not None:
            return self._asynsqs

        c = self._asynsqs = AsyncSQSConnection(
            sqs_connection=self.sqs(queue=queue),
            region=self.region,
            fetch_message_attributes=self.fetch_message_attributes,
        )
        return c

    @property
    def conninfo(self):
        return self.connection.client

    @property
    def transport_options(self):
        return self.connection.client.transport_options

    @cached_property
    def visibility_timeout(self):
        return (self.transport_options.get('visibility_timeout') or
                self.default_visibility_timeout)

    @cached_property
    def predefined_queues(self):
        """Map of queue_name to predefined queue settings."""
        return self.transport_options.get('predefined_queues', {})

    @cached_property
    def queue_name_prefix(self):
        return self.transport_options.get('queue_name_prefix', '')

    @cached_property
    def supports_fanout(self):
        return False

    @cached_property
    def region(self):
        return (self.transport_options.get('region') or
                boto3.Session().region_name or
                self.default_region)

    @cached_property
    def regioninfo(self):
        return self.transport_options.get('regioninfo')

    @cached_property
    def is_secure(self):
        return self.transport_options.get('is_secure')

    @cached_property
    def port(self):
        return self.transport_options.get('port')

    @cached_property
    def endpoint_url(self):
        if self.conninfo.hostname is not None:
            scheme = 'https' if self.is_secure else 'http'
            if self.conninfo.port is not None:
                port = f':{self.conninfo.port}'
            else:
                port = ''
            return '{}://{}{}'.format(
                scheme,
                self.conninfo.hostname,
                port
            )

    @cached_property
    def wait_time_seconds(self):
        return self.transport_options.get('wait_time_seconds',
                                          self.default_wait_time_seconds)

    @cached_property
    def sqs_base64_encoding(self):
        return self.transport_options.get('sqs_base64_encoding', True)

    @cached_property
    def fetch_message_attributes(self):
        return self.transport_options.get('fetch_message_attributes')


class Transport(virtual.Transport):
    """SQS Transport.

    Additional queue attributes can be supplied to SQS during queue
    creation by passing an ``sqs-creation-attributes`` key in
    transport_options. ``sqs-creation-attributes`` must be a dict whose
    key-value pairs correspond with Attributes in the
    `CreateQueue SQS API`_.

    For example, to have SQS queues created with server-side encryption
    enabled using the default Amazon Managed Customer Master Key, you
    can set ``KmsMasterKeyId`` Attribute. When the queue is initially
    created by Kombu, encryption will be enabled.

    .. code-block:: python

        from kombu.transport.SQS import Transport

        transport = Transport(
            ...,
            transport_options={
                'sqs-creation-attributes': {
                    'KmsMasterKeyId': 'alias/aws/sqs',
                },
            }
        )

    .. _CreateQueue SQS API: https://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_CreateQueue.html#API_CreateQueue_RequestParameters

    The ``ApproximateReceiveCount`` message attribute is fetched by this
    transport by default. Requested message attributes can be changed by
    setting ``fetch_message_attributes`` in the transport options.

    .. code-block:: python

        from kombu.transport.SQS import Transport

        transport = Transport(
            ...,
            transport_options={
                'fetch_message_attributes': ["All"],
            }
        )

    .. _Message Attributes: https://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_ReceiveMessage.html#SQS-ReceiveMessage-request-AttributeNames

    """  # noqa: E501

    Channel = Channel

    polling_interval = 1
    wait_time_seconds = 0
    default_port = None
    connection_errors = (
        virtual.Transport.connection_errors +
        (exceptions.BotoCoreError, socket.error)
    )
    channel_errors = (
        virtual.Transport.channel_errors + (exceptions.BotoCoreError,)
    )
    driver_type = 'sqs'
    driver_name = 'sqs'

    implements = virtual.Transport.implements.extend(
        asynchronous=True,
        exchange_type=frozenset(['direct']),
    )

    @property
    def default_connection_params(self):
        return {'port': self.default_port}



"""Text Utilities."""
# flake8: noqa


from __future__ import annotations

from difflib import SequenceMatcher
from typing import Iterable, Iterator

from kombu import version_info_t


def escape_regex(p, white=''):
    # type: (str, str) -> str
    """Escape string for use within a regular expression."""
    # what's up with re.escape? that code must be neglected or something
    return ''.join(c if c.isalnum() or c in white
                   else ('\\000' if c == '\000' else '\\' + c)
                   for c in p)


def fmatch_iter(needle: str, haystack: Iterable[str], min_ratio: float = 0.6) -> Iterator[tuple[float, str]]:
    """Fuzzy match: iteratively.

    Yields
    ------
        Tuple: of ratio and key.
    """
    for key in haystack:
        ratio = SequenceMatcher(None, needle, key).ratio()
        if ratio >= min_ratio:
            yield ratio, key


def fmatch_best(needle: str, haystack: Iterable[str], min_ratio: float = 0.6) -> str | None:
    """Fuzzy match - Find best match (scalar)."""
    try:
        return sorted(
            fmatch_iter(needle, haystack, min_ratio), reverse=True,
        )[0][1]
    except IndexError:
        return None


def version_string_as_tuple(s: str) -> version_info_t:
    """Convert version string to version info tuple."""
    v = _unpack_version(*s.split('.'))
    # X.Y.3a1 -> (X, Y, 3, 'a1')
    if isinstance(v.micro, str):
        v = version_info_t(v.major, v.minor, *_splitmicro(*v[2:]))
    # X.Y.3a1-40 -> (X, Y, 3, 'a1', '40')
    if not v.serial and v.releaselevel and '-' in v.releaselevel:
        v = version_info_t(*list(v[0:3]) + v.releaselevel.split('-'))
    return v


def _unpack_version(
    major: str,
    minor: str | int = 0,
    micro: str | int = 0,
    releaselevel: str = '',
    serial: str = ''
) -> version_info_t:
    return version_info_t(int(major), int(minor), micro, releaselevel, serial)


def _splitmicro(micro: str, releaselevel: str = '', serial: str = '') -> tuple[int, str, str]:
    for index, char in enumerate(micro):
        if not char.isdigit():
            break
    else:
        return int(micro or 0), releaselevel, serial
    return int(micro[:index]), micro[index:], serial



from __future__ import annotations

from typing import Any

from kombu.asynchronous.aws.sqs.connection import AsyncSQSConnection


def connect_sqs(
    aws_access_key_id: str | None = None,
    aws_secret_access_key: str | None = None,
    **kwargs: Any
) -> AsyncSQSConnection:
    """Return async connection to Amazon SQS."""
    from .sqs.connection import AsyncSQSConnection
    return AsyncSQSConnection(
        aws_access_key_id, aws_secret_access_key, **kwargs
    )



"""Serialization utilities."""

from __future__ import annotations

import codecs
import os
import pickle
import sys
from collections import namedtuple
from contextlib import contextmanager
from io import BytesIO

from .exceptions import (ContentDisallowed, DecodeError, EncodeError,
                         SerializerNotInstalled, reraise)
from .utils.compat import entrypoints
from .utils.encoding import bytes_to_str, str_to_bytes

__all__ = ('pickle', 'loads', 'dumps', 'register', 'unregister')
SKIP_DECODE = frozenset(['binary', 'ascii-8bit'])
TRUSTED_CONTENT = frozenset(['application/data', 'application/text'])

if sys.platform.startswith('java'):  # pragma: no cover

    def _decode(t, coding):
        return codecs.getdecoder(coding)(t)[0]
else:
    _decode = codecs.decode

pickle_load = pickle.load

#: We have to use protocol 4 until we drop support for Python 3.6 and 3.7.
pickle_protocol = int(os.environ.get('PICKLE_PROTOCOL', 4))

codec = namedtuple('codec', ('content_type', 'content_encoding', 'encoder'))


@contextmanager
def _reraise_errors(wrapper,
                    include=(Exception,), exclude=(SerializerNotInstalled,)):
    try:
        yield
    except exclude:
        raise
    except include as exc:
        reraise(wrapper, wrapper(exc), sys.exc_info()[2])


def pickle_loads(s, load=pickle_load):
    # used to support buffer objects
    return load(BytesIO(s))


def parenthesize_alias(first, second):
    return f'{first} ({second})' if first else second


class SerializerRegistry:
    """The registry keeps track of serialization methods."""

    def __init__(self):
        self._encoders = {}
        self._decoders = {}
        self._default_encode = None
        self._default_content_type = None
        self._default_content_encoding = None
        self._disabled_content_types = set()
        self.type_to_name = {}
        self.name_to_type = {}

    def register(self, name, encoder, decoder, content_type,
                 content_encoding='utf-8'):
        """Register a new encoder/decoder.

        Arguments:
        ---------
            name (str): A convenience name for the serialization method.

            encoder (callable): A method that will be passed a python data
                structure and should return a string representing the
                serialized data.  If :const:`None`, then only a decoder
                will be registered. Encoding will not be possible.

            decoder (Callable): A method that will be passed a string
                representing serialized data and should return a python
                data structure.  If :const:`None`, then only an encoder
                will be registered.  Decoding will not be possible.

            content_type (str): The mime-type describing the serialized
                structure.

            content_encoding (str): The content encoding (character set) that
                the `decoder` method will be returning. Will usually be
                `utf-8`, `us-ascii`, or `binary`.
        """
        if encoder:
            self._encoders[name] = codec(
                content_type, content_encoding, encoder,
            )
        if decoder:
            self._decoders[content_type] = decoder
        self.type_to_name[content_type] = name
        self.name_to_type[name] = content_type

    def enable(self, name):
        if '/' not in name:
            name = self.name_to_type[name]
        self._disabled_content_types.discard(name)

    def disable(self, name):
        if '/' not in name:
            name = self.name_to_type[name]
        self._disabled_content_types.add(name)

    def unregister(self, name):
        """Unregister registered encoder/decoder.

        Arguments:
        ---------
            name (str): Registered serialization method name.

        Raises
        ------
            SerializerNotInstalled: If a serializer by that name
                cannot be found.
        """
        try:
            content_type = self.name_to_type[name]
            self._decoders.pop(content_type, None)
            self._encoders.pop(name, None)
            self.type_to_name.pop(content_type, None)
            self.name_to_type.pop(name, None)
        except KeyError:
            raise SerializerNotInstalled(
                f'No encoder/decoder installed for {name}')

    def _set_default_serializer(self, name):
        """Set the default serialization method used by this library.

        Arguments:
        ---------
            name (str): The name of the registered serialization method.
                For example, `json` (default), `pickle`, `yaml`, `msgpack`,
                or any custom methods registered using :meth:`register`.

        Raises
        ------
            SerializerNotInstalled: If the serialization method
                requested is not available.
        """
        try:
            (self._default_content_type, self._default_content_encoding,
             self._default_encode) = self._encoders[name]
        except KeyError:
            raise SerializerNotInstalled(
                f'No encoder installed for {name}')

    def dumps(self, data, serializer=None):
        """Encode data.

        Serialize a data structure into a string suitable for sending
        as an AMQP message body.

        Arguments:
        ---------
            data (List, Dict, str): The message data to send.

            serializer (str): An optional string representing
                the serialization method you want the data marshalled
                into. (For example, `json`, `raw`, or `pickle`).

                If :const:`None` (default), then json will be used, unless
                `data` is a :class:`str` or :class:`unicode` object. In this
                latter case, no serialization occurs as it would be
                unnecessary.

                Note that if `serializer` is specified, then that
                serialization method will be used even if a :class:`str`
                or :class:`unicode` object is passed in.

        Returns
        -------
            Tuple[str, str, str]: A three-item tuple containing the
            content type (e.g., `application/json`), content encoding, (e.g.,
            `utf-8`) and a string containing the serialized data.

        Raises
        ------
            SerializerNotInstalled: If the serialization method
                requested is not available.
        """
        if serializer == 'raw':
            return raw_encode(data)
        if serializer and not self._encoders.get(serializer):
            raise SerializerNotInstalled(
                f'No encoder installed for {serializer}')

        # If a raw string was sent, assume binary encoding
        # (it's likely either ASCII or a raw binary file, and a character
        # set of 'binary' will encompass both, even if not ideal.
        if not serializer and isinstance(data, bytes):
            # In Python 3+, this would be "bytes"; allow binary data to be
            # sent as a message without getting encoder errors
            return 'application/data', 'binary', data

        # For Unicode objects, force it into a string
        if not serializer and isinstance(data, str):
            with _reraise_errors(EncodeError, exclude=()):
                payload = data.encode('utf-8')
            return 'text/plain', 'utf-8', payload

        if serializer:
            content_type, content_encoding, encoder = \
                self._encoders[serializer]
        else:
            encoder = self._default_encode
            content_type = self._default_content_type
            content_encoding = self._default_content_encoding

        with _reraise_errors(EncodeError):
            payload = encoder(data)
        return content_type, content_encoding, payload

    def loads(self, data, content_type, content_encoding,
              accept=None, force=False, _trusted_content=TRUSTED_CONTENT):
        """Decode serialized data.

        Deserialize a data stream as serialized using `dumps`
        based on `content_type`.

        Arguments:
        ---------
            data (bytes, buffer, str): The message data to deserialize.

            content_type (str): The content-type of the data.
                (e.g., `application/json`).

            content_encoding (str): The content-encoding of the data.
                (e.g., `utf-8`, `binary`, or `us-ascii`).

            accept (Set): List of content-types to accept.

        Raises
        ------
            ContentDisallowed: If the content-type is not accepted.

        Returns
        -------
            Any: The unserialized data.
        """
        content_type = (bytes_to_str(content_type) if content_type
                        else 'application/data')
        if accept is not None:
            if content_type not in _trusted_content \
                    and content_type not in accept:
                raise self._for_untrusted_content(content_type, 'untrusted')
        else:
            if content_type in self._disabled_content_types and not force:
                raise self._for_untrusted_content(content_type, 'disabled')
        content_encoding = (content_encoding or 'utf-8').lower()

        if data:
            decode = self._decoders.get(content_type)
            if decode:
                with _reraise_errors(DecodeError):
                    return decode(data)
            if content_encoding not in SKIP_DECODE and \
                    not isinstance(data, str):
                with _reraise_errors(DecodeError):
                    return _decode(data, content_encoding)
        return data

    def _for_untrusted_content(self, ctype, why):
        return ContentDisallowed(
            'Refusing to deserialize {} content of type {}'.format(
                why,
                parenthesize_alias(self.type_to_name.get(ctype, ctype), ctype),
            ),
        )


#: Global registry of serializers/deserializers.
registry = SerializerRegistry()
dumps = registry.dumps
loads = registry.loads
register = registry.register
unregister = registry.unregister


def raw_encode(data):
    """Special case serializer."""
    content_type = 'application/data'
    payload = data
    if isinstance(payload, str):
        content_encoding = 'utf-8'
        with _reraise_errors(EncodeError, exclude=()):
            payload = payload.encode(content_encoding)
    else:
        content_encoding = 'binary'
    return content_type, content_encoding, payload


def register_json():
    """Register a encoder/decoder for JSON serialization."""
    from kombu.utils import json as _json

    registry.register('json', _json.dumps, _json.loads,
                      content_type='application/json',
                      content_encoding='utf-8')


def register_yaml():
    """Register a encoder/decoder for YAML serialization.

    It is slower than JSON, but allows for more data types
    to be serialized. Useful if you need to send data such as dates

    """
    try:
        import yaml
        registry.register('yaml', yaml.safe_dump, yaml.safe_load,
                          content_type='application/x-yaml',
                          content_encoding='utf-8')
    except ImportError:

        def not_available(*args, **kwargs):
            """Raise SerializerNotInstalled.

            Used in case a client receives a yaml message, but yaml
            isn't installed.
            """
            raise SerializerNotInstalled(
                'No decoder installed for YAML. Install the PyYAML library')
        registry.register('yaml', None, not_available, 'application/x-yaml')


def unpickle(s):
    return pickle_loads(str_to_bytes(s))


def register_pickle():
    """Register pickle serializer.

    The fastest serialization method, but restricts
    you to python clients.
    """
    def pickle_dumps(obj, dumper=pickle.dumps):
        return dumper(obj, protocol=pickle_protocol)

    registry.register('pickle', pickle_dumps, unpickle,
                      content_type='application/x-python-serialize',
                      content_encoding='binary')


def register_msgpack():
    """Register msgpack serializer.

    See Also
    --------
        https://msgpack.org/.
    """
    pack = unpack = None
    try:
        import msgpack
        if msgpack.version >= (0, 4):
            from msgpack import packb, unpackb

            def pack(s):  # noqa
                return packb(s, use_bin_type=True)

            def unpack(s):  # noqa
                return unpackb(s, raw=False)
        else:
            def version_mismatch(*args, **kwargs):
                raise SerializerNotInstalled(
                    'msgpack requires msgpack-python >= 0.4.0')
            pack = unpack = version_mismatch
    except (ImportError, ValueError):
        def not_available(*args, **kwargs):
            raise SerializerNotInstalled(
                'No decoder installed for msgpack. '
                'Please install the msgpack-python library')
        pack = unpack = not_available
    registry.register(
        'msgpack', pack, unpack,
        content_type='application/x-msgpack',
        content_encoding='binary',
    )


# Register the base serialization methods.
register_json()
register_pickle()
register_yaml()
register_msgpack()

# Default serializer is 'json'
registry._set_default_serializer('json')

NOTSET = object()


def enable_insecure_serializers(choices=NOTSET):
    """Enable serializers that are considered to be unsafe.

    Note:
    ----
        Will enable ``pickle``, ``yaml`` and ``msgpack`` by default, but you
        can also specify a list of serializers (by name or content type)
        to enable.
    """
    choices = ['pickle', 'yaml', 'msgpack'] if choices is NOTSET else choices
    if choices is not None:
        for choice in choices:
            try:
                registry.enable(choice)
            except KeyError:
                pass


def disable_insecure_serializers(allowed=NOTSET):
    """Disable untrusted serializers.

    Will disable all serializers except ``json``
    or you can specify a list of deserializers to allow.

    Note:
    ----
        Producers will still be able to serialize data
        in these formats, but consumers will not accept
        incoming data using the untrusted content types.
    """
    allowed = ['json'] if allowed is NOTSET else allowed
    for name in registry._decoders:
        registry.disable(name)
    if allowed is not None:
        for name in allowed:
            registry.enable(name)


# Insecure serializers are disabled by default since v3.0
disable_insecure_serializers()

# Load entrypoints from installed extensions
for ep, args in entrypoints('kombu.serializers'):  # pragma: no cover
    register(ep.name, *args)


def prepare_accept_content(content_types, name_to_type=None):
    """Replace aliases of content_types with full names from registry.

    Raises
    ------
        SerializerNotInstalled: If the serialization method
            requested is not available.
    """
    name_to_type = registry.name_to_type if not name_to_type else name_to_type
    if content_types is not None:
        try:
            return {n if '/' in n else name_to_type[n] for n in content_types}
        except KeyError as e:
            raise SerializerNotInstalled(
                f'No encoder/decoder installed for {e.args[0]}')
    return content_types



from __future__ import annotations

from .base import (AbstractChannel, Base64, BrokerState, Channel, Empty,
                   Management, Message, NotEquivalentError, QoS, Transport,
                   UndeliverableWarning, binding_key_t, queue_binding_t)

__all__ = (
    'Base64', 'NotEquivalentError', 'UndeliverableWarning', 'BrokerState',
    'QoS', 'Message', 'AbstractChannel', 'Channel', 'Management', 'Transport',
    'Empty', 'binding_key_t', 'queue_binding_t',
)



"""GCP Pub/Sub transport module for kombu.

More information about GCP Pub/Sub:
https://cloud.google.com/pubsub

Features
========
* Type: Virtual
* Supports Direct: Yes
* Supports Topic: No
* Supports Fanout: Yes
* Supports Priority: No
* Supports TTL: No

Connection String
=================

Connection string has the following formats:

.. code-block::

    gcpubsub://projects/project-name

Transport Options
=================
* ``queue_name_prefix``: (str) Prefix for queue names.
* ``ack_deadline_seconds``: (int) The maximum time after receiving a message
  and acknowledging it before pub/sub redelivers the message.
* ``expiration_seconds``: (int) Subscriptions without any subscriber
  activity or changes made to their properties are removed after this period.
  Examples of subscriber activities include open connections,
  active pulls, or successful pushes.
* ``wait_time_seconds``: (int) The maximum time to wait for new messages.
  Defaults to 10.
* ``retry_timeout_seconds``: (int) The maximum time to wait before retrying.
* ``bulk_max_messages``: (int) The maximum number of messages to pull in bulk.
  Defaults to 32.
"""

from __future__ import annotations

import dataclasses
import datetime
import string
import threading
from concurrent.futures import (FIRST_COMPLETED, Future, ThreadPoolExecutor,
                                wait)
from contextlib import suppress
from os import getpid
from queue import Empty
from threading import Lock
from time import monotonic, sleep
from uuid import NAMESPACE_OID, uuid3

from _socket import gethostname
from _socket import timeout as socket_timeout
from google.api_core.exceptions import (AlreadyExists, DeadlineExceeded,
                                        PermissionDenied)
from google.api_core.retry import Retry
from google.cloud import monitoring_v3
from google.cloud.monitoring_v3 import query
from google.cloud.pubsub_v1 import PublisherClient, SubscriberClient
from google.cloud.pubsub_v1 import exceptions as pubsub_exceptions
from google.cloud.pubsub_v1.publisher import exceptions as publisher_exceptions
from google.cloud.pubsub_v1.subscriber import \
    exceptions as subscriber_exceptions
from google.pubsub_v1 import gapic_version as package_version

from kombu.entity import TRANSIENT_DELIVERY_MODE
from kombu.log import get_logger
from kombu.utils.encoding import bytes_to_str, safe_str
from kombu.utils.json import dumps, loads
from kombu.utils.objects import cached_property

from . import virtual

logger = get_logger('kombu.transport.gcpubsub')

# dots are replaced by dash, all other punctuation replaced by underscore.
PUNCTUATIONS_TO_REPLACE = set(string.punctuation) - {'_', '.', '-'}
CHARS_REPLACE_TABLE = {
    ord('.'): ord('-'),
    **{ord(c): ord('_') for c in PUNCTUATIONS_TO_REPLACE},
}


class UnackedIds:
    """Threadsafe list of ack_ids."""

    def __init__(self):
        self._list = []
        self._lock = Lock()

    def append(self, val):
        # append is atomic
        self._list.append(val)

    def extend(self, vals: list):
        # extend is atomic
        self._list.extend(vals)

    def pop(self, index=-1):
        with self._lock:
            return self._list.pop(index)

    def remove(self, val):
        with self._lock, suppress(ValueError):
            self._list.remove(val)

    def __len__(self):
        with self._lock:
            return len(self._list)

    def __getitem__(self, item):
        # getitem is atomic
        return self._list[item]


class AtomicCounter:
    """Threadsafe counter.

    Returns the value after inc/dec operations.
    """

    def __init__(self, initial=0):
        self._value = initial
        self._lock = Lock()

    def inc(self, n=1):
        with self._lock:
            self._value += n
            return self._value

    def dec(self, n=1):
        with self._lock:
            self._value -= n
            return self._value

    def get(self):
        with self._lock:
            return self._value


@dataclasses.dataclass
class QueueDescriptor:
    """Pub/Sub queue descriptor."""

    name: str
    topic_path: str  # projects/{project_id}/topics/{topic_id}
    subscription_id: str
    subscription_path: str  # projects/{project_id}/subscriptions/{subscription_id}
    unacked_ids: UnackedIds = dataclasses.field(default_factory=UnackedIds)


class Channel(virtual.Channel):
    """GCP Pub/Sub channel."""

    supports_fanout = True
    do_restore = False  # pub/sub does that for us
    default_wait_time_seconds = 10
    default_ack_deadline_seconds = 240
    default_expiration_seconds = 86400
    default_retry_timeout_seconds = 300
    default_bulk_max_messages = 32

    _min_ack_deadline = 10
    _fanout_exchanges = set()
    _unacked_extender: threading.Thread = None
    _stop_extender = threading.Event()
    _n_channels = AtomicCounter()
    _queue_cache: dict[str, QueueDescriptor] = {}
    _tmp_subscriptions: set[str] = set()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.pool = ThreadPoolExecutor()
        logger.info('new GCP pub/sub channel: %s', self.conninfo.hostname)

        self.project_id = Transport.parse_uri(self.conninfo.hostname)
        if self._n_channels.inc() == 1:
            Channel._unacked_extender = threading.Thread(
                target=self._extend_unacked_deadline,
                daemon=True,
            )
            self._stop_extender.clear()
            Channel._unacked_extender.start()

    def entity_name(self, name: str, table=CHARS_REPLACE_TABLE) -> str:
        """Format AMQP queue name into a valid Pub/Sub queue name."""
        if not name.startswith(self.queue_name_prefix):
            name = self.queue_name_prefix + name

        return str(safe_str(name)).translate(table)

    def _queue_bind(self, exchange, routing_key, pattern, queue):
        exchange_type = self.typeof(exchange).type
        queue = self.entity_name(queue)
        logger.debug(
            'binding queue: %s to %s exchange: %s with routing_key: %s',
            queue,
            exchange_type,
            exchange,
            routing_key,
        )

        filter_args = {}
        if exchange_type == 'direct':
            # Direct exchange is implemented as a single subscription
            # E.g. for exchange 'test_direct':
            # -topic:'test_direct'
            #  -bound queue:'direct1':
            #  -subscription: direct1' on topic 'test_direct'
            #   -filter:routing_key'
            filter_args = {
                'filter': f'attributes.routing_key="{routing_key}"'
            }
            subscription_path = self.subscriber.subscription_path(
                self.project_id, queue
            )
            message_retention_duration = self.expiration_seconds
        elif exchange_type == 'fanout':
            # Fanout exchange is implemented as a separate subscription.
            # E.g. for exchange 'test_fanout':
            # -topic:'test_fanout'
            #  -bound queue 'fanout1':
            #    -subscription:'fanout1-uuid' on topic 'test_fanout'
            #  -bound queue 'fanout2':
            #    -subscription:'fanout2-uuid' on topic 'test_fanout'
            uid = f'{uuid3(NAMESPACE_OID, f"{gethostname()}.{getpid()}")}'
            uniq_sub_name = f'{queue}-{uid}'
            subscription_path = self.subscriber.subscription_path(
                self.project_id, uniq_sub_name
            )
            self._tmp_subscriptions.add(subscription_path)
            self._fanout_exchanges.add(exchange)
            message_retention_duration = 600
        else:
            raise NotImplementedError(
                f'exchange type {exchange_type} not implemented'
            )
        exchange_topic = self._create_topic(
            self.project_id, exchange, message_retention_duration
        )
        self._create_subscription(
            topic_path=exchange_topic,
            subscription_path=subscription_path,
            filter_args=filter_args,
            msg_retention=message_retention_duration,
        )
        qdesc = QueueDescriptor(
            name=queue,
            topic_path=exchange_topic,
            subscription_id=queue,
            subscription_path=subscription_path,
        )
        self._queue_cache[queue] = qdesc

    def _create_topic(
        self,
        project_id: str,
        topic_id: str,
        message_retention_duration: int = None,
    ) -> str:
        topic_path = self.publisher.topic_path(project_id, topic_id)
        if self._is_topic_exists(topic_path):
            # topic creation takes a while, so skip if possible
            logger.debug('topic: %s exists', topic_path)
            return topic_path
        try:
            logger.debug('creating topic: %s', topic_path)
            request = {'name': topic_path}
            if message_retention_duration:
                request[
                    'message_retention_duration'
                ] = f'{message_retention_duration}s'
            self.publisher.create_topic(request=request)
        except AlreadyExists:
            pass

        return topic_path

    def _is_topic_exists(self, topic_path: str) -> bool:
        topics = self.publisher.list_topics(
            request={"project": f'projects/{self.project_id}'}
        )
        for t in topics:
            if t.name == topic_path:
                return True
        return False

    def _create_subscription(
        self,
        project_id: str = None,
        topic_id: str = None,
        topic_path: str = None,
        subscription_path: str = None,
        filter_args=None,
        msg_retention: int = None,
    ) -> str:
        subscription_path = (
            subscription_path
            or self.subscriber.subscription_path(self.project_id, topic_id)
        )
        topic_path = topic_path or self.publisher.topic_path(
            project_id, topic_id
        )
        try:
            logger.debug(
                'creating subscription: %s, topic: %s, filter: %s',
                subscription_path,
                topic_path,
                filter_args,
            )
            msg_retention = msg_retention or self.expiration_seconds
            self.subscriber.create_subscription(
                request={
                    "name": subscription_path,
                    "topic": topic_path,
                    'ack_deadline_seconds': self.ack_deadline_seconds,
                    'expiration_policy': {
                        'ttl': f'{self.expiration_seconds}s'
                    },
                    'message_retention_duration': f'{msg_retention}s',
                    **(filter_args or {}),
                }
            )
        except AlreadyExists:
            pass
        return subscription_path

    def _delete(self, queue, *args, **kwargs):
        """Delete a queue by name."""
        queue = self.entity_name(queue)
        logger.info('deleting queue: %s', queue)
        qdesc = self._queue_cache.get(queue)
        if not qdesc:
            return
        self.subscriber.delete_subscription(
            request={"subscription": qdesc.subscription_path}
        )
        self._queue_cache.pop(queue, None)

    def _put(self, queue, message, **kwargs):
        """Put a message onto the queue."""
        queue = self.entity_name(queue)
        qdesc = self._queue_cache[queue]
        routing_key = self._get_routing_key(message)
        logger.debug(
            'putting message to queue: %s, topic: %s, routing_key: %s',
            queue,
            qdesc.topic_path,
            routing_key,
        )
        encoded_message = dumps(message)
        self.publisher.publish(
            qdesc.topic_path,
            encoded_message.encode("utf-8"),
            routing_key=routing_key,
        )

    def _put_fanout(self, exchange, message, routing_key, **kwargs):
        """Put a message onto fanout exchange."""
        self._lookup(exchange, routing_key)
        topic_path = self.publisher.topic_path(self.project_id, exchange)
        logger.debug(
            'putting msg to fanout exchange: %s, topic: %s',
            exchange,
            topic_path,
        )
        encoded_message = dumps(message)
        self.publisher.publish(
            topic_path,
            encoded_message.encode("utf-8"),
            retry=Retry(deadline=self.retry_timeout_seconds),
        )

    def _get(self, queue: str, timeout: float = None):
        """Retrieves a single message from a queue."""
        queue = self.entity_name(queue)
        qdesc = self._queue_cache[queue]
        try:
            response = self.subscriber.pull(
                request={
                    'subscription': qdesc.subscription_path,
                    'max_messages': 1,
                },
                retry=Retry(deadline=self.retry_timeout_seconds),
                timeout=timeout or self.wait_time_seconds,
            )
        except DeadlineExceeded:
            raise Empty()

        if len(response.received_messages) == 0:
            raise Empty()

        message = response.received_messages[0]
        ack_id = message.ack_id
        payload = loads(message.message.data)
        delivery_info = payload['properties']['delivery_info']
        logger.debug(
            'queue:%s got message, ack_id: %s, payload: %s',
            queue,
            ack_id,
            payload['properties'],
        )
        if self._is_auto_ack(payload['properties']):
            logger.debug('auto acking message ack_id: %s', ack_id)
            self._do_ack([ack_id], qdesc.subscription_path)
        else:
            delivery_info['gcpubsub_message'] = {
                'queue': queue,
                'ack_id': ack_id,
                'message_id': message.message.message_id,
                'subscription_path': qdesc.subscription_path,
            }
            qdesc.unacked_ids.append(ack_id)

        return payload

    def _is_auto_ack(self, payload_properties: dict):
        exchange = payload_properties['delivery_info']['exchange']
        delivery_mode = payload_properties['delivery_mode']
        return (
            delivery_mode == TRANSIENT_DELIVERY_MODE
            or exchange in self._fanout_exchanges
        )

    def _get_bulk(self, queue: str, timeout: float):
        """Retrieves bulk of messages from a queue."""
        prefixed_queue = self.entity_name(queue)
        qdesc = self._queue_cache[prefixed_queue]
        max_messages = self._get_max_messages_estimate()
        if not max_messages:
            raise Empty()
        try:
            response = self.subscriber.pull(
                request={
                    'subscription': qdesc.subscription_path,
                    'max_messages': max_messages,
                },
                retry=Retry(deadline=self.retry_timeout_seconds),
                timeout=timeout or self.wait_time_seconds,
            )
        except DeadlineExceeded:
            raise Empty()

        received_messages = response.received_messages
        if len(received_messages) == 0:
            raise Empty()

        auto_ack_ids = []
        ret_payloads = []
        logger.debug(
            'batching %d messages from queue: %s',
            len(received_messages),
            prefixed_queue,
        )
        for message in received_messages:
            ack_id = message.ack_id
            payload = loads(bytes_to_str(message.message.data))
            delivery_info = payload['properties']['delivery_info']
            delivery_info['gcpubsub_message'] = {
                'queue': prefixed_queue,
                'ack_id': ack_id,
                'message_id': message.message.message_id,
                'subscription_path': qdesc.subscription_path,
            }
            if self._is_auto_ack(payload['properties']):
                auto_ack_ids.append(ack_id)
            else:
                qdesc.unacked_ids.append(ack_id)
            ret_payloads.append(payload)
        if auto_ack_ids:
            logger.debug('auto acking ack_ids: %s', auto_ack_ids)
            self._do_ack(auto_ack_ids, qdesc.subscription_path)

        return queue, ret_payloads

    def _get_max_messages_estimate(self) -> int:
        max_allowed = self.qos.can_consume_max_estimate()
        max_if_unlimited = self.bulk_max_messages
        return max_if_unlimited if max_allowed is None else max_allowed

    def _lookup(self, exchange, routing_key, default=None):
        exchange_info = self.state.exchanges.get(exchange, {})
        if not exchange_info:
            return super()._lookup(exchange, routing_key, default)
        ret = self.typeof(exchange).lookup(
            self.get_table(exchange),
            exchange,
            routing_key,
            default,
        )
        if ret:
            return ret
        logger.debug(
            'no queues bound to exchange: %s, binding on the fly',
            exchange,
        )
        self.queue_bind(exchange, exchange, routing_key)
        return [exchange]

    def _size(self, queue: str) -> int:
        """Return the number of messages in a queue.

        This is a *rough* estimation, as Pub/Sub doesn't provide
        an exact API.
        """
        queue = self.entity_name(queue)
        if queue not in self._queue_cache:
            return 0
        qdesc = self._queue_cache[queue]
        result = query.Query(
            self.monitor,
            self.project_id,
            'pubsub.googleapis.com/subscription/num_undelivered_messages',
            end_time=datetime.datetime.now(),
            minutes=1,
        ).select_resources(subscription_id=qdesc.subscription_id)

        # monitoring API requires the caller to have the monitoring.viewer
        # role. Since we can live without the exact number of messages
        # in the queue, we can ignore the exception and allow users to
        # use the transport without this role.
        with suppress(PermissionDenied):
            return sum(
                content.points[0].value.int64_value for content in result
            )
        return -1

    def basic_ack(self, delivery_tag, multiple=False):
        """Acknowledge one message."""
        if multiple:
            raise NotImplementedError('multiple acks not implemented')

        delivery_info = self.qos.get(delivery_tag).delivery_info
        pubsub_message = delivery_info['gcpubsub_message']
        ack_id = pubsub_message['ack_id']
        queue = pubsub_message['queue']
        logger.debug('ack message. queue: %s ack_id: %s', queue, ack_id)
        subscription_path = pubsub_message['subscription_path']
        self._do_ack([ack_id], subscription_path)
        qdesc = self._queue_cache[queue]
        qdesc.unacked_ids.remove(ack_id)
        super().basic_ack(delivery_tag)

    def _do_ack(self, ack_ids: list[str], subscription_path: str):
        self.subscriber.acknowledge(
            request={"subscription": subscription_path, "ack_ids": ack_ids},
            retry=Retry(deadline=self.retry_timeout_seconds),
        )

    def _purge(self, queue: str):
        """Delete all current messages in a queue."""
        queue = self.entity_name(queue)
        qdesc = self._queue_cache.get(queue)
        if not qdesc:
            return

        n = self._size(queue)
        self.subscriber.seek(
            request={
                "subscription": qdesc.subscription_path,
                "time": datetime.datetime.now(),
            }
        )
        return n

    def _extend_unacked_deadline(self):
        thread_id = threading.get_native_id()
        logger.info(
            'unacked deadline extension thread: [%s] started',
            thread_id,
        )
        min_deadline_sleep = self._min_ack_deadline / 2
        sleep_time = max(min_deadline_sleep, self.ack_deadline_seconds / 4)
        while not self._stop_extender.wait(sleep_time):
            for qdesc in self._queue_cache.values():
                if len(qdesc.unacked_ids) == 0:
                    logger.debug(
                        'thread [%s]: no unacked messages for %s',
                        thread_id,
                        qdesc.subscription_path,
                    )
                    continue
                logger.debug(
                    'thread [%s]: extend ack deadline for %s: %d msgs [%s]',
                    thread_id,
                    qdesc.subscription_path,
                    len(qdesc.unacked_ids),
                    list(qdesc.unacked_ids),
                )
                self.subscriber.modify_ack_deadline(
                    request={
                        "subscription": qdesc.subscription_path,
                        "ack_ids": list(qdesc.unacked_ids),
                        "ack_deadline_seconds": self.ack_deadline_seconds,
                    }
                )
        logger.info(
            'unacked deadline extension thread [%s] stopped', thread_id
        )

    def after_reply_message_received(self, queue: str):
        queue = self.entity_name(queue)
        sub = self.subscriber.subscription_path(self.project_id, queue)
        logger.debug(
            'after_reply_message_received: queue: %s, sub: %s', queue, sub
        )
        self._tmp_subscriptions.add(sub)

    @cached_property
    def subscriber(self):
        return SubscriberClient()

    @cached_property
    def publisher(self):
        return PublisherClient()

    @cached_property
    def monitor(self):
        return monitoring_v3.MetricServiceClient()

    @property
    def conninfo(self):
        return self.connection.client

    @property
    def transport_options(self):
        return self.connection.client.transport_options

    @cached_property
    def wait_time_seconds(self):
        return self.transport_options.get(
            'wait_time_seconds', self.default_wait_time_seconds
        )

    @cached_property
    def retry_timeout_seconds(self):
        return self.transport_options.get(
            'retry_timeout_seconds', self.default_retry_timeout_seconds
        )

    @cached_property
    def ack_deadline_seconds(self):
        return self.transport_options.get(
            'ack_deadline_seconds', self.default_ack_deadline_seconds
        )

    @cached_property
    def queue_name_prefix(self):
        return self.transport_options.get('queue_name_prefix', 'kombu-')

    @cached_property
    def expiration_seconds(self):
        return self.transport_options.get(
            'expiration_seconds', self.default_expiration_seconds
        )

    @cached_property
    def bulk_max_messages(self):
        return self.transport_options.get(
            'bulk_max_messages', self.default_bulk_max_messages
        )

    def close(self):
        """Close the channel."""
        logger.debug('closing channel')
        while self._tmp_subscriptions:
            sub = self._tmp_subscriptions.pop()
            with suppress(Exception):
                logger.debug('deleting subscription: %s', sub)
                self.subscriber.delete_subscription(
                    request={"subscription": sub}
                )
        if not self._n_channels.dec():
            self._stop_extender.set()
            Channel._unacked_extender.join()
        super().close()

    @staticmethod
    def _get_routing_key(message):
        routing_key = (
            message['properties']
            .get('delivery_info', {})
            .get('routing_key', '')
        )
        return routing_key


class Transport(virtual.Transport):
    """GCP Pub/Sub transport."""

    Channel = Channel

    can_parse_url = True
    polling_interval = 0.1
    connection_errors = virtual.Transport.connection_errors + (
        pubsub_exceptions.TimeoutError,
    )
    channel_errors = (
        virtual.Transport.channel_errors
        + (
            publisher_exceptions.FlowControlLimitError,
            publisher_exceptions.MessageTooLargeError,
            publisher_exceptions.PublishError,
            publisher_exceptions.TimeoutError,
            publisher_exceptions.PublishToPausedOrderingKeyException,
        )
        + (subscriber_exceptions.AcknowledgeError,)
    )

    driver_type = 'gcpubsub'
    driver_name = 'pubsub_v1'

    implements = virtual.Transport.implements.extend(
        exchange_type=frozenset(['direct', 'fanout']),
    )

    def __init__(self, client, **kwargs):
        super().__init__(client, **kwargs)
        self._pool = ThreadPoolExecutor()
        self._get_bulk_future_to_queue: dict[Future, str] = dict()

    def driver_version(self):
        return package_version.__version__

    @staticmethod
    def parse_uri(uri: str) -> str:
        # URL like:
        #  gcpubsub://projects/project-name

        project = uri.split('gcpubsub://projects/')[1]
        return project.strip('/')

    @classmethod
    def as_uri(self, uri: str, include_password=False, mask='**') -> str:
        return uri or 'gcpubsub://'

    def drain_events(self, connection, timeout=None):
        time_start = monotonic()
        polling_interval = self.polling_interval
        if timeout and polling_interval and polling_interval > timeout:
            polling_interval = timeout
        while 1:
            try:
                self._drain_from_active_queues(timeout=timeout)
            except Empty:
                if timeout and monotonic() - time_start >= timeout:
                    raise socket_timeout()
                if polling_interval:
                    sleep(polling_interval)
            else:
                break

    def _drain_from_active_queues(self, timeout):
        # cleanup empty requests from prev run
        self._rm_empty_bulk_requests()

        # submit new requests for all active queues
        # longer timeout means less frequent polling
        # and more messages in a single bulk
        self._submit_get_bulk_requests(timeout=10)

        done, _ = wait(
            self._get_bulk_future_to_queue,
            timeout=timeout,
            return_when=FIRST_COMPLETED,
        )
        empty = {f for f in done if f.exception()}
        done -= empty
        for f in empty:
            self._get_bulk_future_to_queue.pop(f, None)

        if not done:
            raise Empty()

        logger.debug('got %d done get_bulk tasks', len(done))
        for f in done:
            queue, payloads = f.result()
            for payload in payloads:
                logger.debug('consuming message from queue: %s', queue)
                if queue not in self._callbacks:
                    logger.warning(
                        'Message for queue %s without consumers', queue
                    )
                    continue
                self._deliver(payload, queue)
            self._get_bulk_future_to_queue.pop(f, None)

    def _rm_empty_bulk_requests(self):
        empty = {
            f
            for f in self._get_bulk_future_to_queue
            if f.done() and f.exception()
        }
        for f in empty:
            self._get_bulk_future_to_queue.pop(f, None)

    def _submit_get_bulk_requests(self, timeout):
        queues_with_submitted_get_bulk = set(
            self._get_bulk_future_to_queue.values()
        )

        for channel in self.channels:
            for queue in channel._active_queues:
                if queue in queues_with_submitted_get_bulk:
                    continue
                future = self._pool.submit(channel._get_bulk, queue, timeout)
                self._get_bulk_future_to_queue[future] = queue



"""Built-in transports."""

from __future__ import annotations

from kombu.utils.compat import _detect_environment
from kombu.utils.imports import symbol_by_name


def supports_librabbitmq() -> bool | None:
    """Return true if :pypi:`librabbitmq` can be used."""
    if _detect_environment() == 'default':
        try:
            import librabbitmq  # noqa
        except ImportError:  # pragma: no cover
            pass
        else:                # pragma: no cover
            return True
    return None


TRANSPORT_ALIASES = {
    'amqp': 'kombu.transport.pyamqp:Transport',
    'amqps': 'kombu.transport.pyamqp:SSLTransport',
    'pyamqp': 'kombu.transport.pyamqp:Transport',
    'librabbitmq': 'kombu.transport.librabbitmq:Transport',
    'confluentkafka': 'kombu.transport.confluentkafka:Transport',
    'kafka': 'kombu.transport.confluentkafka:Transport',
    'memory': 'kombu.transport.memory:Transport',
    'redis': 'kombu.transport.redis:Transport',
    'rediss': 'kombu.transport.redis:Transport',
    'SQS': 'kombu.transport.SQS:Transport',
    'sqs': 'kombu.transport.SQS:Transport',
    'mongodb': 'kombu.transport.mongodb:Transport',
    'zookeeper': 'kombu.transport.zookeeper:Transport',
    'sqlalchemy': 'kombu.transport.sqlalchemy:Transport',
    'sqla': 'kombu.transport.sqlalchemy:Transport',
    'SLMQ': 'kombu.transport.SLMQ.Transport',
    'slmq': 'kombu.transport.SLMQ.Transport',
    'filesystem': 'kombu.transport.filesystem:Transport',
    'qpid': 'kombu.transport.qpid:Transport',
    'sentinel': 'kombu.transport.redis:SentinelTransport',
    'consul': 'kombu.transport.consul:Transport',
    'etcd': 'kombu.transport.etcd:Transport',
    'azurestoragequeues': 'kombu.transport.azurestoragequeues:Transport',
    'azureservicebus': 'kombu.transport.azureservicebus:Transport',
    'pyro': 'kombu.transport.pyro:Transport',
    'gcpubsub': 'kombu.transport.gcpubsub:Transport',
}

_transport_cache = {}


def resolve_transport(transport: str | None = None) -> str | None:
    """Get transport by name.

    Arguments:
    ---------
        transport (Union[str, type]): This can be either
            an actual transport class, or the fully qualified
            path to a transport class, or the alias of a transport.
    """
    if isinstance(transport, str):
        try:
            transport = TRANSPORT_ALIASES[transport]
        except KeyError:
            if '.' not in transport and ':' not in transport:
                from kombu.utils.text import fmatch_best
                alt = fmatch_best(transport, TRANSPORT_ALIASES)
                if alt:
                    raise KeyError(
                        'No such transport: {}.  Did you mean {}?'.format(
                            transport, alt))
                raise KeyError(f'No such transport: {transport}')
        else:
            if callable(transport):
                transport = transport()
        return symbol_by_name(transport)
    return transport


def get_transport_cls(transport: str | None = None) -> str | None:
    """Get transport class by name.

    The transport string is the full path to a transport class, e.g.::

        "kombu.transport.pyamqp:Transport"

    If the name does not include `"."` (is not fully qualified),
    the alias table will be consulted.
    """
    if transport not in _transport_cache:
        _transport_cache[transport] = resolve_transport(transport)
    return _transport_cache[transport]



"""Generic resource pool implementation."""

from __future__ import annotations

import os
from collections import deque
from queue import Empty
from queue import LifoQueue as _LifoQueue
from typing import TYPE_CHECKING

from . import exceptions
from .utils.compat import register_after_fork
from .utils.functional import lazy

if TYPE_CHECKING:
    from types import TracebackType


def _after_fork_cleanup_resource(resource):
    try:
        resource.force_close_all()
    except Exception:
        pass


class LifoQueue(_LifoQueue):
    """Last in first out version of Queue."""

    def _init(self, maxsize):
        self.queue = deque()


class Resource:
    """Pool of resources."""

    LimitExceeded = exceptions.LimitExceeded

    close_after_fork = False

    def __init__(self, limit=None, preload=None, close_after_fork=None):
        self._limit = limit
        self.preload = preload or 0
        self._closed = False
        self.close_after_fork = (
            close_after_fork
            if close_after_fork is not None else self.close_after_fork
        )

        self._resource = LifoQueue()
        self._dirty = set()
        if self.close_after_fork and register_after_fork is not None:
            register_after_fork(self, _after_fork_cleanup_resource)
        self.setup()

    def setup(self):
        raise NotImplementedError('subclass responsibility')

    def _add_when_empty(self):
        if self.limit and len(self._dirty) >= self.limit:
            raise self.LimitExceeded(self.limit)
        # All taken, put new on the queue and
        # try get again, this way the first in line
        # will get the resource.
        self._resource.put_nowait(self.new())

    def acquire(self, block=False, timeout=None):
        """Acquire resource.

        Arguments:
        ---------
            block (bool): If the limit is exceeded,
                then block until there is an available item.
            timeout (float): Timeout to wait
                if ``block`` is true.  Default is :const:`None` (forever).

        Raises
        ------
            LimitExceeded: if block is false and the limit has been exceeded.
        """
        if self._closed:
            raise RuntimeError('Acquire on closed pool')
        if self.limit:
            while 1:
                try:
                    R = self._resource.get(block=block, timeout=timeout)
                except Empty:
                    self._add_when_empty()
                else:
                    try:
                        R = self.prepare(R)
                    except BaseException:
                        if isinstance(R, lazy):
                            # not evaluated yet, just put it back
                            self._resource.put_nowait(R)
                        else:
                            # evaluated so must try to release/close first.
                            self.release(R)
                        raise
                    self._dirty.add(R)
                    break
        else:
            R = self.prepare(self.new())

        def release():
            """Release resource so it can be used by another thread.

            Warnings:
            --------
                The caller is responsible for discarding the object,
                and to never use the resource again.  A new resource must
                be acquired if so needed.
            """
            self.release(R)
        R.release = release

        return R

    def prepare(self, resource):
        return resource

    def close_resource(self, resource):
        resource.close()

    def release_resource(self, resource):
        pass

    def replace(self, resource):
        """Replace existing resource with a new instance.

        This can be used in case of defective resources.
        """
        if self.limit:
            self._dirty.discard(resource)
        self.close_resource(resource)

    def release(self, resource):
        if self.limit:
            self._dirty.discard(resource)
            self._resource.put_nowait(resource)
            self.release_resource(resource)
        else:
            self.close_resource(resource)

    def collect_resource(self, resource):
        pass

    def force_close_all(self, close_pool=True):
        """Close and remove all resources in the pool (also those in use).

        Used to close resources from parent processes after fork
        (e.g. sockets/connections).

        Arguments:
        ---------
            close_pool (bool): If True (default) then the pool is marked
                as closed. In case of False the pool can be reused.
        """
        if self._closed:
            return
        self._closed = close_pool
        dirty = self._dirty
        resource = self._resource
        while 1:  # - acquired
            try:
                dres = dirty.pop()
            except KeyError:
                break
            try:
                self.collect_resource(dres)
            except AttributeError:  # Issue #78
                pass
        while 1:  # - available
            # deque supports '.clear', but lists do not, so for that
            # reason we use pop here, so that the underlying object can
            # be any object supporting '.pop' and '.append'.
            try:
                res = resource.queue.pop()
            except IndexError:
                break
            try:
                self.collect_resource(res)
            except AttributeError:
                pass  # Issue #78

    def resize(self, limit, force=False, ignore_errors=False, reset=False):
        prev_limit = self._limit
        if (self._dirty and 0 < limit < self._limit) and not ignore_errors:
            if not force:
                raise RuntimeError(
                    "Can't shrink pool when in use: was={} now={}".format(
                        self._limit, limit))
            reset = True
        self._limit = limit
        if reset:
            try:
                self.force_close_all(close_pool=False)
            except Exception:
                pass
        self.setup()
        if limit < prev_limit:
            self._shrink_down(collect=limit > 0)

    def _shrink_down(self, collect=True):
        class Noop:
            def __enter__(self):
                pass

            def __exit__(
                self,
                exc_type: type,
                exc_val: Exception,
                exc_tb: TracebackType
            ) -> None:
                pass

        resource = self._resource
        # Items to the left are last recently used, so we remove those first.
        with getattr(resource, 'mutex', Noop()):
            # keep in mind the dirty resources are not shrinking
            while len(resource.queue) and \
                    (len(resource.queue) + len(self._dirty)) > self.limit:
                R = resource.queue.popleft()
                if collect:
                    self.collect_resource(R)

    @property
    def limit(self):
        return self._limit

    @limit.setter
    def limit(self, limit):
        self.resize(limit)

    if os.environ.get('KOMBU_DEBUG_POOL'):  # pragma: no cover
        _orig_acquire = acquire
        _orig_release = release

        _next_resource_id = 0

        def acquire(self, *args, **kwargs):
            import traceback
            id = self._next_resource_id = self._next_resource_id + 1
            print(f'+{id} ACQUIRE {self.__class__.__name__}')
            r = self._orig_acquire(*args, **kwargs)
            r._resource_id = id
            print(f'-{id} ACQUIRE {self.__class__.__name__}')
            if not hasattr(r, 'acquired_by'):
                r.acquired_by = []
            r.acquired_by.append(traceback.format_stack())
            return r

        def release(self, resource):
            id = resource._resource_id
            print(f'+{id} RELEASE {self.__class__.__name__}')
            r = self._orig_release(resource)
            print(f'-{id} RELEASE {self.__class__.__name__}')
            self._next_resource_id -= 1
            return r



"""Amazon SQS queue implementation."""

from __future__ import annotations

from vine import transform

from .message import AsyncMessage

_all__ = ['AsyncQueue']


def list_first(rs):
    """Get the first item in a list, or None if list empty."""
    return rs[0] if len(rs) == 1 else None


class AsyncQueue:
    """Async SQS Queue."""

    def __init__(self, connection=None, url=None, message_class=AsyncMessage):
        self.connection = connection
        self.url = url
        self.message_class = message_class
        self.visibility_timeout = None

    def _NA(self, *args, **kwargs):
        raise NotImplementedError()
    count_slow = dump = save_to_file = save_to_filename = save = \
        save_to_s3 = load_from_s3 = load_from_file = load_from_filename = \
        load = clear = _NA

    def get_attributes(self, attributes='All', callback=None):
        return self.connection.get_queue_attributes(
            self, attributes, callback,
        )

    def set_attribute(self, attribute, value, callback=None):
        return self.connection.set_queue_attribute(
            self, attribute, value, callback,
        )

    def get_timeout(self, callback=None, _attr='VisibilityTimeout'):
        return self.get_attributes(
            _attr, transform(
                self._coerce_field_value, callback, _attr, int,
            ),
        )

    def _coerce_field_value(self, key, type, response):
        return type(response[key])

    def set_timeout(self, visibility_timeout, callback=None):
        return self.set_attribute(
            'VisibilityTimeout', visibility_timeout,
            transform(
                self._on_timeout_set, callback,
            )
        )

    def _on_timeout_set(self, visibility_timeout):
        if visibility_timeout:
            self.visibility_timeout = visibility_timeout
        return self.visibility_timeout

    def add_permission(self, label, aws_account_id, action_name,
                       callback=None):
        return self.connection.add_permission(
            self, label, aws_account_id, action_name, callback,
        )

    def remove_permission(self, label, callback=None):
        return self.connection.remove_permission(self, label, callback)

    def read(self, visibility_timeout=None, wait_time_seconds=None,
             callback=None):
        return self.get_messages(
            1, visibility_timeout,
            wait_time_seconds=wait_time_seconds,
            callback=transform(list_first, callback),
        )

    def write(self, message, delay_seconds=None, callback=None):
        return self.connection.send_message(
            self, message.get_body_encoded(), delay_seconds,
            callback=transform(self._on_message_sent, callback, message),
        )

    def write_batch(self, messages, callback=None):
        return self.connection.send_message_batch(
            self, messages, callback=callback,
        )

    def _on_message_sent(self, orig_message, new_message):
        orig_message.id = new_message.id
        orig_message.md5 = new_message.md5
        return new_message

    def get_messages(self, num_messages=1, visibility_timeout=None,
                     attributes=None, wait_time_seconds=None, callback=None):
        return self.connection.receive_message(
            self, number_messages=num_messages,
            visibility_timeout=visibility_timeout,
            attributes=attributes,
            wait_time_seconds=wait_time_seconds,
            callback=callback,
        )

    def delete_message(self, message, callback=None):
        return self.connection.delete_message(self, message, callback)

    def delete_message_batch(self, messages, callback=None):
        return self.connection.delete_message_batch(
            self, messages, callback=callback,
        )

    def change_message_visibility_batch(self, messages, callback=None):
        return self.connection.change_message_visibility_batch(
            self, messages, callback=callback,
        )

    def delete(self, callback=None):
        return self.connection.delete_queue(self, callback=callback)

    def count(self, page_size=10, vtimeout=10, callback=None,
              _attr='ApproximateNumberOfMessages'):
        return self.get_attributes(
            _attr, callback=transform(
                self._coerce_field_value, callback, _attr, int,
            ),
        )



from __future__ import annotations

from collections import deque
from io import BytesIO

import urllib3

from kombu.asynchronous.hub import Hub, get_event_loop
from kombu.exceptions import HttpError

from .base import BaseClient, Request

__all__ = ('Urllib3Client',)

from ...utils.encoding import bytes_to_str

DEFAULT_USER_AGENT = 'Mozilla/5.0 (compatible; urllib3)'
EXTRA_METHODS = frozenset(['DELETE', 'OPTIONS', 'PATCH'])


def _get_pool_key_parts(request: Request) -> list[str]:
    _pool_key_parts = []

    if request.network_interface:
        _pool_key_parts.append(f"interface={request.network_interface}")

    if request.validate_cert:
        _pool_key_parts.append("validate_cert=True")
    else:
        _pool_key_parts.append("validate_cert=False")

    if request.ca_certs:
        _pool_key_parts.append(f"ca_certs={request.ca_certs}")

    if request.client_cert:
        _pool_key_parts.append(f"client_cert={request.client_cert}")

    if request.client_key:
        _pool_key_parts.append(f"client_key={request.client_key}")

    return _pool_key_parts


class Urllib3Client(BaseClient):
    """Urllib3 HTTP Client."""

    _pools = {}

    def __init__(self, hub: Hub | None = None, max_clients: int = 10):
        hub = hub or get_event_loop()
        super().__init__(hub)
        self.max_clients = max_clients
        self._pending = deque()
        self._timeout_check_tref = self.hub.call_repeatedly(
            1.0, self._timeout_check,
        )

    def pools_close(self):
        for pool in self._pools.values():
            pool.close()
        self._pools.clear()

    def close(self):
        self._timeout_check_tref.cancel()
        self.pools_close()

    def add_request(self, request):
        self._pending.append(request)
        self._process_queue()
        return request

    def get_pool(self, request: Request):
        _pool_key_parts = _get_pool_key_parts(request=request)

        _proxy_url = None
        proxy_headers = None
        if request.proxy_host:
            _proxy_url = urllib3.util.Url(
                scheme=None,
                host=request.proxy_host,
                port=request.proxy_port,
            )
            if request.proxy_username:
                proxy_headers = urllib3.make_headers(
                    proxy_basic_auth=(
                        f"{request.proxy_username}"
                        f":{request.proxy_password}"
                    )
                )
            else:
                proxy_headers = None

            _proxy_url = _proxy_url.url

            _pool_key_parts.append(f"proxy={_proxy_url}")
            if proxy_headers:
                _pool_key_parts.append(f"proxy_headers={str(proxy_headers)}")

        _pool_key = "|".join(_pool_key_parts)
        if _pool_key in self._pools:
            return self._pools[_pool_key]

        # create new pool
        if _proxy_url:
            _pool = urllib3.ProxyManager(
                proxy_url=_proxy_url,
                num_pools=self.max_clients,
                proxy_headers=proxy_headers
            )
        else:
            _pool = urllib3.PoolManager(num_pools=self.max_clients)

        # Network Interface
        if request.network_interface:
            _pool.connection_pool_kw['source_address'] = (
                request.network_interface,
                0
            )

        # SSL Verification
        if request.validate_cert:
            _pool.connection_pool_kw['cert_reqs'] = 'CERT_REQUIRED'
        else:
            _pool.connection_pool_kw['cert_reqs'] = 'CERT_NONE'

        # CA Certificates
        if request.ca_certs is not None:
            _pool.connection_pool_kw['ca_certs'] = request.ca_certs
        elif request.validate_cert is True:
            try:
                from certifi import where
                _pool.connection_pool_kw['ca_certs'] = where()
            except ImportError:
                pass

        # Client Certificates
        if request.client_cert is not None:
            _pool.connection_pool_kw['cert_file'] = request.client_cert
        if request.client_key is not None:
            _pool.connection_pool_kw['key_file'] = request.client_key

        self._pools[_pool_key] = _pool

        return _pool

    def _timeout_check(self):
        self._process_pending_requests()

    def _process_pending_requests(self):
        while self._pending:
            request = self._pending.popleft()
            self._process_request(request)

    def _process_request(self, request: Request):
        # Prepare headers
        headers = request.headers
        headers.setdefault(
            'User-Agent',
            bytes_to_str(request.user_agent or DEFAULT_USER_AGENT)
        )
        headers.setdefault(
            'Accept-Encoding',
            'gzip,deflate' if request.use_gzip else 'none'
        )

        # Authentication
        if request.auth_username is not None:
            headers.update(
                urllib3.util.make_headers(
                    basic_auth=(
                        f"{request.auth_username}"
                        f":{request.auth_password or ''}"
                    )
                )
            )

        # Make the request using urllib3
        try:
            _pool = self.get_pool(request=request)
            response = _pool.request(
                request.method,
                request.url,
                headers=headers,
                body=request.body,
                preload_content=False,
                redirect=request.follow_redirects,
            )
            buffer = BytesIO(response.data)
            response_obj = self.Response(
                request=request,
                code=response.status,
                headers=response.headers,
                buffer=buffer,
                effective_url=response.geturl(),
                error=None
            )
        except urllib3.exceptions.HTTPError as e:
            response_obj = self.Response(
                request=request,
                code=599,
                headers={},
                buffer=None,
                effective_url=None,
                error=HttpError(599, str(e))
            )

        request.on_ready(response_obj)

    def _process_queue(self):
        self._process_pending_requests()

    def on_readable(self, fd):
        pass

    def on_writable(self, fd):
        pass

    def _setup_request(self, curl, request, buffer, headers):
        pass
